#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <ORBitservices/orbit-imodule.h>

#include "php_orbit.h"
#include "phorb.h"

gboolean phorb_load_idl_from_file(const char *file)
{
	phorbIDL *idl;
	int ret;

	idl = g_hash_table_lookup(OrbSG(idl_cache), file);

	if (idl) {
		struct stat statbuf;
		time_t now = time(NULL);

		if (lstat(file, &statbuf) == SUCCESS)
			ret = statbuf.st_ctime < idl->lastchanged;
		else
			ret = FALSE;

		if (!ret) {
			g_hash_table_remove(OrbSG(idl_cache), file);
			idl = NULL;
		} else
			idl->lastchecked = now;
	}

	if (idl)
		ret = TRUE;
	else {
		CORBA_sequence_ORBit_IInterface *ifaces;
		CORBA_sequence_CORBA_TypeCode *types;

		idl = phorb_alloc_idl(file);

		phorb_setup_typecodes(idl);
		phorb_setup_exceptions(idl);

		ifaces = ORBit_iinterfaces_from_file(file,
		                                     OrbSG(idl_cpp_arguments),
		                                     &types);
		if (ifaces) {
			CORBA_unsigned_long i;

			for(i = 0; i < ifaces->_length; i++) {
				ORBit_IInterface *iface = &ifaces->_buffer[i];

				if (phorb_find_interface_description(idl,
				    	iface->tc->repo_id))
					continue;

				phorb_init_interface(idl, iface);
			}

			for(i = 0; i < types->_length; i++) {
				CORBA_TypeCode tc = types->_buffer[i];

				phorb_store_typecode(idl, tc);
			}

			CORBA_sequence_ORBit_IInterface__freekids(ifaces,
			                                          NULL);
			CORBA_free(ifaces);
			CORBA_sequence_CORBA_TypeCode__freekids(types, NULL);
			CORBA_free(types);

			ret = TRUE;
			g_hash_table_insert(OrbSG(idl_cache), idl->name, idl);
		} else
			ret = FALSE;
	}

	if (ret) {
		phorb_xref_interfaces(idl);

		OrbSG(idls) = g_slist_prepend(OrbSG(idls), idl);
	} else if (idl)
		phorb_free_idl(idl);

	return ret;
}

static void corba_free(gpointer value)
{
	CORBA_Object_release((CORBA_Object)value, NULL);
}

phorbIDL *phorb_alloc_idl(const char *name) {
	phorbIDL *idl = g_new(phorbIDL, 1);
	time_t now = time(NULL);

	idl->interfaces = g_hash_table_new_full(g_str_hash, g_str_equal, NULL,
	                  	(GDestroyNotify)phorb_free_interface_info);
	idl->typecodes = g_hash_table_new_full(g_str_hash, g_str_equal,
	                                       NULL, corba_free);
	idl->exceptions = g_hash_table_new_full(g_str_hash, g_str_equal,
	                                        g_free, g_free);

	idl->lastchanged = now;
	idl->lastaccessed = now;
	idl->lastchecked = now;

	idl->name = g_strdup(name);

	return idl;
}

void phorb_free_idl(phorbIDL *idl) {
	g_hash_table_destroy(idl->interfaces);
	idl->interfaces = NULL;

	g_hash_table_destroy(idl->typecodes);
	idl->typecodes = NULL;

	g_hash_table_destroy(idl->exceptions);
	idl->exceptions = NULL;

	g_free(idl->name);
	idl->name = NULL;

	g_free(idl);
}

static gboolean phorb_check_idl(gchar *fname, phorbIDL *idl, gpointer data) {
	struct stat statbuf;
	time_t now = time(NULL);

	if (idl->lastaccessed + OrbSG(idl_cache_timeout) < now)
		return TRUE;
	else if (idl->lastchecked + OrbSG(idl_cache_check) >= now)
		return FALSE;
	else if (stat(idl->name, &statbuf) != SUCCESS)
		return TRUE;
	else if (statbuf.st_ctime > idl->lastchanged)
		return TRUE;

	idl->lastchecked = now;

	return FALSE;
}

void phorb_check_idl_cache(void)
{
	g_hash_table_foreach_remove(OrbSG(idl_cache),
	                           (GHRFunc)phorb_check_idl, NULL);
}
