#include "php.h"
#include "zend_compile.h"
#include "php_ini.h"
#include "php_globals.h"
#include "php_orbit.h"
#include "phorb.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>


extern const char *ORBit_tk_to_name(CORBA_unsigned_long tk);

#define ALIGN_VAL(buf, align)		*buf = ALIGN_ADDRESS(*buf, align)
#define GET_VAL(buf, type)		(*(type *)(*buf))
#define ADVANCE_PTR(buf, offset)	*buf = ((guchar *)(*buf)) + offset

gboolean phorb_put_zval(gconstpointer *buf, CORBA_TypeCode tc, zval *val);

static gboolean put_short(gconstpointer *buf, zval *val) {
	CORBA_short v;

	convert_to_long(val);
	v = Z_LVAL_P(val);
	if (v != Z_LVAL_P(val)) {
		zend_error(E_WARNING, "%s out of range",
		           ORBit_tk_to_name(CORBA_tk_short));
		return FALSE;
	}

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_SHORT);
	GET_VAL(buf, CORBA_short) = v;
	ADVANCE_PTR(buf, sizeof v);

	return TRUE;
}

static gboolean put_ushort(gconstpointer *buf, zval *val)
{
	CORBA_unsigned_short v;

	convert_to_long(val);
	v = Z_LVAL_P(val);
	if (v != Z_LVAL_P(val)) {
		zend_error(E_WARNING, "%s out of range",
		           ORBit_tk_to_name(CORBA_tk_ushort));
		return FALSE;
	}

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_SHORT);
	GET_VAL(buf, CORBA_unsigned_short) = v;
	ADVANCE_PTR(buf, sizeof v);

	return TRUE;
}

static gboolean put_long(gconstpointer *buf, zval *val)
{
	CORBA_long v;

	convert_to_long(val);
	v = Z_LVAL_P(val);

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG);
	GET_VAL(buf, CORBA_long) = v;
	ADVANCE_PTR(buf, sizeof v);

	return TRUE;
}

static gboolean put_ulong(gconstpointer *buf, zval *val)
{
	CORBA_unsigned_long v;

	if (Z_TYPE_P(val) == IS_DOUBLE) {
		if (Z_DVAL_P(val) < 0 || Z_DVAL_P(val) > (double)G_MAXUINT32) {
			zend_error(E_WARNING, "%s out of range",
			           ORBit_tk_to_name(CORBA_tk_ulong));
			return FALSE;
		}
		v = (CORBA_unsigned_long)Z_DVAL_P(val);
	} else {
		convert_to_long(val);
		if (Z_LVAL_P(val) < 0) {
			zend_error(E_WARNING, "%s out of range",
			           ORBit_tk_to_name(CORBA_tk_ulong));
			return FALSE;
		}
		v = Z_LVAL_P(val);
	}

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG);
	GET_VAL(buf, CORBA_long) = v;
	ADVANCE_PTR(buf, sizeof v);

	return TRUE;
}

static gboolean put_longlong(gconstpointer *buf, zval *val)
{
	CORBA_long_long v;

	if (Z_TYPE_P(val) == IS_DOUBLE) {
		if (Z_DVAL_P(val) < (double)G_MININT64
		    || Z_DVAL_P(val) > (double)G_MAXINT64) {
			zend_error(E_WARNING, "%s out of range",
			           ORBit_tk_to_name(CORBA_tk_longlong));
			return FALSE;
		}
		v = (CORBA_long_long)Z_DVAL_P(val);
	} else {
		convert_to_long(val);
		v = Z_LVAL_P(val);
	}

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG_LONG);
	GET_VAL(buf, CORBA_long_long) = v;
	ADVANCE_PTR(buf, sizeof v);

	return CORBA_TRUE;
}

static gboolean put_ulonglong(gconstpointer *buf, zval *val)
{
	CORBA_unsigned_long_long v;

	if (Z_TYPE_P(val) == IS_DOUBLE) {
		if (Z_DVAL_P(val) < 0 || Z_DVAL_P(val) > (double)G_MAXUINT64) {
			zend_error(E_WARNING, "%s out of range",
			           ORBit_tk_to_name(CORBA_tk_ulonglong));
			return FALSE;
		}
		v = (CORBA_unsigned_long_long)Z_DVAL_P(val);
	} else {
		convert_to_long(val);
		if (Z_LVAL_P(val) < 0) {
			zend_error(E_WARNING, "%s out of range",
			           ORBit_tk_to_name(CORBA_tk_ulonglong));
			return FALSE;
		}
		v = Z_LVAL_P(val);
	}

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG_LONG);
	GET_VAL(buf, CORBA_unsigned_long_long) = v;
	ADVANCE_PTR(buf, sizeof v);

	return CORBA_TRUE;
}

static gboolean put_octet(gconstpointer *buf, zval *val)
{
	CORBA_octet v;

	convert_to_long(val);
	v = Z_LVAL_P(val);
	if (v != Z_LVAL_P(val)) {
		zend_error(E_WARNING, "%s out of range",
		           ORBit_tk_to_name(CORBA_tk_octet));
		return FALSE;
	}

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_OCTET);
	GET_VAL(buf, CORBA_octet) = v;
	ADVANCE_PTR(buf, sizeof v);

	return TRUE;
}

static gboolean put_boolean(gconstpointer *buf, zval *val)
{
	CORBA_boolean v;

	convert_to_boolean(val);
	v = Z_LVAL_P(val) != 0;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_BOOLEAN);
	GET_VAL(buf, CORBA_boolean) = v;
	ADVANCE_PTR(buf, sizeof v);

	return TRUE;
}

static gboolean put_float(gconstpointer *buf, zval *val)
{
	CORBA_float v;

	convert_to_double(val);
	v = Z_DVAL_P(val);

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_FLOAT);
	GET_VAL(buf, CORBA_float) = v;
	ADVANCE_PTR(buf, sizeof v);

	return TRUE;
}

static gboolean put_double(gconstpointer *buf, zval *val)
{
	CORBA_double v;

	convert_to_double(val);
	v = Z_DVAL_P(val);

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_DOUBLE);
	GET_VAL(buf, CORBA_double) = v;
	ADVANCE_PTR(buf, sizeof v);

	return TRUE;
}

static gboolean put_long_double(gconstpointer *buf, zval *val)
{
	CORBA_long_double v;

	convert_to_double(val);
	v = Z_DVAL_P(val);

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG_DOUBLE);
	GET_VAL(buf, CORBA_long_double) = v;
	ADVANCE_PTR(buf, sizeof v);

	return TRUE;
}

static gboolean put_char(gconstpointer *buf, zval *val)
{
	CORBA_char *v;

	convert_to_string(val);
	v = Z_STRVAL_P(val);
	if (Z_STRLEN_P(val) != 1) {
		zend_error(E_WARNING, "%s must have length = 1",
		           ORBit_tk_to_name(CORBA_tk_char));
		return FALSE;
	}

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_CHAR);
	GET_VAL(buf, CORBA_char) = v[0];
	ADVANCE_PTR(buf, sizeof v[0]);

	return TRUE;
}

static gboolean put_wchar(gconstpointer *buf, zval *val)
{
	GError *error = NULL;
	gunichar2 *utf16;
	glong len = 0;

	convert_to_string(val);
	if (!Z_STRVAL_P(val)) {
		zend_error(E_WARNING, "Can't convert to string");
		return FALSE;
	}

	utf16 = g_utf8_to_utf16(Z_STRVAL_P(val), -1, NULL, &len, &error);
	if (!utf16) {
		zend_error(E_WARNING, "UTF8 read error: %s", error->message);
		g_error_free(error);
		return FALSE;
	}

	if (len != 1) {
		zend_error(E_WARNING, "%s must have length = 1",
		           ORBit_tk_to_name(CORBA_tk_wchar));
		g_free(utf16);
		return FALSE;
	}

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_WCHAR);
	GET_VAL(buf, CORBA_wchar) = utf16[0];
	ADVANCE_PTR(buf, sizeof(CORBA_wchar));

	g_free(utf16);

	return TRUE;
}

static gboolean put_enum(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	CORBA_long v;

	v = phorb_enum_find_member(tc, val);
	if (v < 0) {
		zend_error(E_WARNING, "Invalid enumeration value");
		return FALSE;
	}

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG);
	GET_VAL(buf, CORBA_unsigned_long) = (CORBA_unsigned_long)v;
	ADVANCE_PTR(buf, sizeof v);

	return TRUE;
}

static gboolean put_struct(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	HashPosition pos;
	ulong index;
	int i;

	ALIGN_VAL(buf, tc->c_align);

	switch(Z_TYPE_P(val)) {
	case IS_ARRAY:
		if (tc->sub_parts < zend_hash_num_elements(HASH_OF(val))) {
			zend_error(E_WARNING,
			           "Too many array members (should be %d)",
			           tc->sub_parts);
			return FALSE;
		}

		zend_hash_internal_pointer_reset_ex(HASH_OF(val), &pos);
		for(i = 0; i < tc->sub_parts;
		    zend_hash_move_forward_ex(HASH_OF(val), &pos), i++) {
			int j;
			char *key;
			zval **item;

			if (zend_hash_get_current_key_ex(HASH_OF(val),
			    	&key, NULL, &index, FALSE, &pos)
			    == HASH_KEY_NON_EXISTANT) {
				zend_error(E_WARNING,
				           "Missing structure member `%s'",
				           tc->subnames[i]);
				return FALSE;
			}
			zend_hash_get_current_data_ex(HASH_OF(val),
			                              (void *)&item, &pos);
			if (!phorb_put_zval(buf, tc->subtypes[i], *item))
				return FALSE;
		}
		return TRUE;
	case IS_OBJECT:
		for(i = 0; i < tc->sub_parts; i++) {
			zval **item;

			if (zend_hash_find(HASH_OF(val),
			    	(char *)tc->subnames[i],
			    	strlen(tc->subnames[i]) + 1,
			    	(void *)&item)
			    == FAILURE) {
				zend_error(E_WARNING,
				           "Missing structure member `%s'",
				           tc->subnames[i]);
					return FALSE;
			}
			if (!phorb_put_zval(buf, tc->subtypes[i], *item))
				return FALSE;
		}
		return TRUE;
	default:
		zend_error(E_WARNING, "Array or Object expected");
		return FALSE;
	}
}

static gboolean put_sequence(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	CORBA_sequence_CORBA_octet *sval;
	CORBA_unsigned_long len;
	gconstpointer seqval;
	HashPosition pos;
	ulong index;
	int i;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_SEQ);
	sval = (CORBA_sequence_CORBA_octet *)(*buf);
	ADVANCE_PTR(buf, sizeof(CORBA_sequence_CORBA_octet));

	if (Z_TYPE_P(val) != IS_ARRAY) {
		if (CORBA_TypeCode_equal(tc->subtypes[0], TC_CORBA_octet,
		                         NULL)) {
			convert_to_string(val);
			len = Z_STRLEN_P(val);

			sval->_buffer = ORBit_alloc_tcval(TC_CORBA_octet, len);
			sval->_length = len;
			sval->_release = CORBA_TRUE;
			memcpy(sval->_buffer, Z_STRVAL_P(val), len);

			return TRUE;
		} else {
			zend_error(E_WARNING, "Array expected");
			return FALSE;
		}
	}

	len = zend_hash_num_elements(HASH_OF(val));

	if (tc->length != 0 && len > tc->length) {
		zend_error(E_WARNING,
		           "Sequence length (%d) exceeds bound (%d)",
		           len, tc->length);
		return FALSE;
	}

	sval->_buffer = ORBit_alloc_tcval(tc->subtypes[0], len);
	sval->_length = len;
	sval->_release = CORBA_TRUE;

	seqval = sval->_buffer;
	zend_hash_internal_pointer_reset_ex(HASH_OF(val), &pos);
	for(i = 0; i < len;
	    zend_hash_move_forward_ex(HASH_OF(val), &pos), i++) {
		int j;
		char *key;
		zval **item;

		if (zend_hash_get_current_key_ex(HASH_OF(val), &key, NULL,
		                                 &index, FALSE, &pos)
		    == HASH_KEY_NON_EXISTANT) {
			zend_error(E_WARNING,
			           "Missing sequence member `%s'",
			           tc->subnames[i]);
			return FALSE;
		}
		zend_hash_get_current_data_ex(HASH_OF(val),
		                              (void **)&item, &pos);
		if (!phorb_put_zval(&seqval, tc->subtypes[0], *item))
			return FALSE;
	}

	return CORBA_TRUE;
}

static gboolean put_array(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	HashPosition pos;
	ulong index;
	int i, align = tc->subtypes[0]->c_align;

	if (Z_TYPE_P(val) != IS_ARRAY) {
		zend_error(E_WARNING, "Array expected");
		return FALSE;
	}

	if (zend_hash_num_elements(HASH_OF(val)) != tc->length) {
		zend_error(E_WARNING,
		           "Array argument should be of length %d, is %d",
		           tc->length,
		           zend_hash_num_elements(HASH_OF(val)));
		return FALSE;
	}

	zend_hash_internal_pointer_reset_ex(HASH_OF(val),&pos);
	for(i = 0; i < tc->length;
	    zend_hash_move_forward_ex(HASH_OF(val), &pos), i++) {
		int j;
		char *key;
		zval **item;

		if (zend_hash_get_current_key_ex(HASH_OF(val), &key, NULL,
		                                 &index, FALSE, &pos)
		    == HASH_KEY_NON_EXISTANT) {
			zend_error(E_WARNING, "Missing array member");
			return FALSE;
		}
		zend_hash_get_current_data_ex(HASH_OF(val),
		                              (void **)&item, &pos);
		ALIGN_VAL(buf, align);
		if (!phorb_put_zval(buf, tc->subtypes[0], *item))
			return FALSE;
	}

	return TRUE;
}

static gboolean put_objref(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	CORBA_Object obj = CORBA_OBJECT_NIL;

	switch(Z_TYPE_P(val)) {
	case IS_BOOL:
	case IS_LONG:
	case IS_DOUBLE:
	case IS_STRING:
		convert_to_long(val);
		if (Z_LVAL_P(val)) {
			zend_error(E_WARNING,
			           "Bad argument type for Object reference");
			return FALSE;
		}
	case IS_NULL:
		val = NULL;
	case IS_OBJECT:
		if (!val || phorb_is_object(val))
			break;
	default:
		zend_error(E_WARNING,
		           "Bad argument type for Object reference");
		return FALSE;
	}

	if (val) {
		CORBA_char *repoid;
		zval **handle;
		int type;

		zend_hash_index_find(Z_OBJPROP_P(val),
		                     0, (void *)&handle);
		obj = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle), &type);

		repoid = phorb_get_object_repoid(obj, NULL);
		if (tc->repo_id[0] && repoid
		    && strcmp(repoid, tc->repo_id) != 0) {
			zend_error(E_WARNING,
			           "Attempt to marshall bad object type "
			           "`%s', should be `%s'",
			           repoid, tc->repo_id);
			CORBA_free(repoid);
			return FALSE;
		}
		CORBA_free(repoid);

		obj = CORBA_Object_duplicate(obj, NULL);
	}

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_POINTER);
	GET_VAL(buf, CORBA_Object) = obj;
	ADVANCE_PTR(buf, sizeof(CORBA_Object));

	return TRUE;
}

static gboolean put_union(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	gconstpointer tmp;
	HashPosition pos;
	CORBA_long arm;
	zval **discriminator, **value;
	int i, sz = 0;

	if (Z_TYPE_P(val) != IS_ARRAY) {
		zend_error(E_WARNING, "Array expected");
		return FALSE;
	}
	if (zend_hash_num_elements(HASH_OF(val)) != 2) {
		zend_error(E_WARNING,
		           "Union arrays must have exactly 2 elements");
		return FALSE;
	}

	zend_hash_internal_pointer_reset_ex(HASH_OF(val), &pos);
	zend_hash_get_current_data_ex(HASH_OF(val), (void *)&discriminator,
	                              &pos);
	zend_hash_move_forward_ex(HASH_OF(val), &pos);
	zend_hash_get_current_data_ex(HASH_OF(val), (void *)&value, &pos);

	ALIGN_VAL(buf, MAX(tc->c_align, tc->discriminator->c_align));
	if (!phorb_put_zval(buf, tc->discriminator, *discriminator))
		return FALSE;

	arm = phorb_union_find_arm(tc, *discriminator);
	if (arm < 0) {
		zend_error(E_WARNING,
		           "discrimator branch does not match any arm "
		           "and there's no default arm");
		return FALSE;
	}

	for(i = 0; i < tc->sub_parts; i++)
		sz = MAX(sz, ORBit_gather_alloc_info(tc->subtypes[i]));

	ALIGN_VAL(buf, tc->c_align);
	tmp = *buf;
	ADVANCE_PTR(buf, sz);

	return phorb_put_zval(&tmp, tc->subtypes[arm], *value);
}

static gboolean put_string(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	convert_to_string(val);
	if (!Z_STRVAL_P(val)) {
		zend_error(E_WARNING, "Can't convert to string");
		return FALSE;
	}

	if (tc->length > 0 && Z_STRLEN_P(val) > tc->length) {
		zend_error(E_WARNING, "String too long");
		return FALSE;
	}
	if (strlen(Z_STRVAL_P(val)) != Z_STRLEN_P(val)) {
		zend_error(E_WARNING, "String may not include embedded nulls");
		return FALSE;
	}

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_POINTER);
	GET_VAL(buf, CORBA_string) = CORBA_string_dup(Z_STRVAL_P(val));
	ADVANCE_PTR(buf, sizeof(CORBA_char *));

	return TRUE;
}

static gboolean put_wstring(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	GError *error = NULL;
	gunichar2 *utf16;
	glong len = 0;

	convert_to_string(val);
	if (!Z_STRVAL_P(val)) {
		zend_error(E_WARNING, "Can't convert to string");
		return FALSE;
	}

	if (strlen(Z_STRVAL_P(val)) != Z_STRLEN_P(val)) {
		zend_error(E_WARNING, "String may not include embedded nulls");
		return FALSE;
	}

	utf16 = g_utf8_to_utf16(Z_STRVAL_P(val), -1, NULL, &len, &error);
	if (!utf16) {
		zend_error(E_WARNING, "UTF8 read error: %s", error->message);
		g_error_free(error);
		return FALSE;
	}

	if (tc->length > 0 && len > tc->length) {
		zend_error(E_WARNING, "String too long");
		g_free(utf16);
		return FALSE;
	}
	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_POINTER);
	GET_VAL(buf, CORBA_wstring) = CORBA_wstring_alloc(len);
	memcpy(GET_VAL(buf, CORBA_wstring), utf16,
	       (len + 1) * sizeof(CORBA_wchar));
	ADVANCE_PTR(buf, sizeof(CORBA_wchar *));

	g_free(utf16);

	return TRUE;
}

static gboolean put_fixed(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	return FALSE;	/* FIXME: implement this */
}

static gboolean put_typecode(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	CORBA_TypeCode obj;
	zval **handle;
	int type;

	if (!phorb_is_typecode(val)) {
		zend_error(E_WARNING, "Attempt to marshall non typecode");
		return FALSE;
	}

	zend_hash_index_find(Z_OBJPROP_P(val), 0, (void *)&handle);
	obj = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(handle), &type);

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_POINTER);
	GET_VAL(buf, CORBA_TypeCode) =
		(CORBA_TypeCode)CORBA_Object_duplicate((CORBA_Object)obj,
		                                       NULL);
	ADVANCE_PTR(buf, sizeof obj);

	return TRUE;
}

static gboolean put_any(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	CORBA_TypeCode obj;
	gconstpointer anyval;
	HashPosition pos;
	CORBA_any *any;
	zval **anytc, *value, **handle;
	int type;

	if (Z_TYPE_P(val) != IS_ARRAY) {
		zend_error(E_WARNING, "Array expected");
		return FALSE;
	}
	if (zend_hash_num_elements(HASH_OF(val)) != 2) {
		zend_error(E_WARNING,
		           "Any arrays must have exactly 2 elements");
		return FALSE;
	}

	zend_hash_internal_pointer_reset_ex(HASH_OF(val), &pos);
	zend_hash_get_current_data_ex(HASH_OF(val), (void *)&anytc, &pos);
	zend_hash_move_forward_ex(HASH_OF(val), &pos);
	zend_hash_get_current_data_ex(HASH_OF(val), (void *)&value, &pos);

	if (!phorb_is_typecode(*anytc)) {
		zend_error(E_WARNING,
		           "Attempt to marshall any without typecode");
		return FALSE;
	}

	zend_hash_index_find(Z_OBJPROP_PP(anytc), 0, (void *)&handle);
	obj = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(handle), &type);

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_ANY);
	any = &GET_VAL(buf, CORBA_any);
	ADVANCE_PTR(buf, sizeof(CORBA_any));

	any->_type = (CORBA_TypeCode)CORBA_Object_duplicate(
	                             	(CORBA_Object)obj, NULL);
	any->_value = ORBit_small_alloc(obj);
	any->_release = CORBA_TRUE;

	anyval = any->_value;
	if (!phorb_put_zval(&anyval, obj, value))
		return FALSE;

	return TRUE;
}

gboolean phorb_put_zval(gconstpointer *buf, CORBA_TypeCode tc, zval *val)
{
	while(tc->kind == CORBA_tk_alias)
		tc = tc->subtypes[0];

	switch(tc->kind) {
	case CORBA_tk_null:
	case CORBA_tk_void:
		return CORBA_TRUE;
	case CORBA_tk_short:
		return put_short(buf, val);
	case CORBA_tk_ushort:
		return put_ushort(buf, val);
	case CORBA_tk_long:
		return put_long(buf, val);
	case CORBA_tk_ulong:
		return put_ulong(buf, val);
	case CORBA_tk_longlong:
		return put_longlong(buf, val);
	case CORBA_tk_ulonglong:
		return put_ulonglong(buf, val);
	case CORBA_tk_float:
		return put_float(buf, val);
	case CORBA_tk_double:
		return put_double(buf, val);
	case CORBA_tk_char:
		return put_char(buf, val);
	case CORBA_tk_wchar:
		return put_wchar(buf, val);
	case CORBA_tk_boolean:
		return put_boolean(buf, val);
	case CORBA_tk_octet:
		return put_octet(buf, val);
	case CORBA_tk_enum:
		return put_enum(buf, tc, val);
	case CORBA_tk_struct:
		return put_struct(buf, tc, val);
	case CORBA_tk_sequence:
		return put_sequence(buf, tc, val);
	case CORBA_tk_objref:
		return put_objref(buf, tc, val);
	case CORBA_tk_union:
		return put_union(buf, tc, val);
	case CORBA_tk_string:
		return put_string(buf, tc, val);
	case CORBA_tk_wstring:
		return put_wstring(buf, tc, val);
	case CORBA_tk_array:
		return put_array(buf, tc, val);
	case CORBA_tk_TypeCode:
		return put_typecode(buf, tc, val);
	case CORBA_tk_any:
		return put_any(buf, tc, val);
	case CORBA_tk_fixed:
		return put_fixed(buf, tc, val);
	case CORBA_tk_except:
		/* we don't marshal exceptions */
	case CORBA_tk_Principal:
	default:
		zend_error(E_WARNING,
		           "Unsupported output typecode `%s' (%d)\n",
		           tc->repo_id, tc->kind);
	return FALSE;
	}
}
