/*
 * php-orbit - a PHP language mapping for the ORBit2 CORBA ORB
 * Copyright (C) 2000-2004  Christophe Saout <christophe@saout.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
                
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "zend_compile.h"
#include "php_ini.h"
#include "php_globals.h"
#include "php_orbit.h"
#include "phorb.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/stat.h>

static zend_class_entry *corba_class_entry;
static zend_class_entry *corbatc_class_entry;

ZEND_DECLARE_MODULE_GLOBALS(orbit)

static int mod_number;

PHP_FUNCTION(orbit_load_idl)
{
	char *fname;
	char full_path_buff[MAXPATHLEN];
	int fname_len;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
	                          &fname, &fname_len) == FAILURE) {
		RETURN_FALSE;
	}

	if (VCWD_REALPATH(fname, full_path_buff)) {
#ifdef ZTS
# ifdef PHP_WIN32
		if (_access(full_path_buff, 0)) {
			RETURN_FALSE;
		}
# else
		if (access(full_path_buff, F_OK)) {
			RETURN_FALSE;
		}
# endif
#endif
	} else {
		RETURN_FALSE;
	}

	if (phorb_load_idl_from_file(full_path_buff)) {
		RETURN_TRUE;
	} else {
		RETURN_FALSE;
	}
}

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_orbit_load_idl, 0, 0, 1)
	ZEND_ARG_INFO(0, idl_file)
ZEND_END_ARG_INFO()

PHP_FUNCTION(orbit_exception_handler)
{
	char *handler;
	int handler_len;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
	                          &handler, &handler_len) == FAILURE)
		RETURN_FALSE;

	phorb_exception_set_user_function(handler);
}

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_orbit_exception_handler, 0, 0, 1)
	ZEND_ARG_INFO(0, handler)
ZEND_END_ARG_INFO()

PHP_METHOD(CORBA, __construct)
{
	CORBA_Environment ev;
	CORBA_Object cobj;
	char *ior;
	int ior_len;
	zval *tmp;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
	                          &ior, &ior_len) == FAILURE) {
		goto error;
	}

	CORBA_exception_init(&ev);
	phorb_ensure_orb(&ev);

	cobj = CORBA_ORB_string_to_object(OrbSG(orb), ior, &ev);
	if (!cobj) {
		zend_error(E_WARNING,
		           "Could not map IOR to Object in CORBA()");
		goto error;
	}

	MAKE_STD_ZVAL(tmp);
	ZVAL_RESOURCE(tmp, phorb_register_object(cobj));

	zend_hash_index_update(Z_OBJPROP_P(this_ptr), 0, &tmp,
	                       sizeof(zval *), NULL);
	zend_update_property_string(zend_get_class_entry(this_ptr TSRMLS_DC),
	                            this_ptr, "_ior", 4, ior);

	CORBA_exception_free(&ev);
	return;

  error:
  	CORBA_exception_free(&ev);
	zval_dtor(this_ptr);
	ZVAL_NULL(this_ptr);
	RETURN_FALSE;
}

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_CORBA___construct, 0, 0, 1)
	ZEND_ARG_INFO(0, ior)
ZEND_END_ARG_INFO()

PHP_METHOD(CORBA, __call)
{
	CORBA_Object cobj;
	HashPosition pos;
	zval **handle, *args, **argv = NULL, **param;
	char *name;
	int i = 0, name_len, type, argc;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sa",
	                          &name, &name_len, &args) == FAILURE) {
		RETURN_FALSE;
	}

	argc = zend_hash_num_elements(Z_ARRVAL_P(args));
	if (argc > 0) {
		argv = safe_emalloc(sizeof(zval *), argc, 0);
		for(zend_hash_internal_pointer_reset_ex(Z_ARRVAL_P(args),
		                                        &pos);
		    zend_hash_get_current_data_ex(Z_ARRVAL_P(args),
		                                  (void **) &param, &pos)
					    			== SUCCESS;
		    zend_hash_move_forward_ex(Z_ARRVAL_P(args), &pos))
			argv[i++] = *param;
	}

	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	cobj = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle), &type);

	phorb_invoke_method(cobj, name, argv, argc, return_value);

	if (argv)
		efree(argv);
}

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_CORBA___call, 0, 0, 2)
	ZEND_ARG_INFO(0, name)
	ZEND_ARG_INFO(0, params)
ZEND_END_ARG_INFO()
											  
PHP_METHOD(CORBA, __get)
{
	CORBA_Object cobj;
	zval **handle;
	char *name;
	int name_len, type;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
	                          &name, &name_len) == FAILURE) {
		RETURN_FALSE;
	}

	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	cobj = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle), &type);

	phorb_get_attribute(cobj, name, return_value);
}

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_CORBA___get, 0, 0, 1)
	ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO()
											  
PHP_METHOD(CORBA, __set)
{
	CORBA_Object cobj;
	zval **handle, *arg;
	char *name;
	int name_len, type;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sz",
	                          &name, &name_len, &arg) == FAILURE) {
		RETURN_FALSE;
	}

	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	cobj = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle), &type);

	if (phorb_set_attribute(cobj, name, arg) == FAILURE) {
		RETURN_FALSE;
	}
}

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_CORBA___set, 0, 0, 2)
	ZEND_ARG_INFO(0, name)
	ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO()
											  
PHP_METHOD(CORBA, _equal)
{
	CORBA_Object cobj, cobj2;
	zval **handle, **handle2, *arg;
	int type;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "O",
	                          &arg, *corba_class_entry) == FAILURE) {
		RETURN_FALSE;
	}

	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	cobj = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle), &type);

	zend_hash_index_find(Z_OBJPROP_P(arg), 0, (void *)&handle2);
	cobj2 = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle2), &type);

	if (cobj == cobj2) {
		RETVAL_TRUE;
	} else {
		RETVAL_FALSE;
	}
}

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_CORBA__equal, 0, 0, 1)
	ZEND_ARG_INFO(0, object2)
ZEND_END_ARG_INFO()

PHP_METHOD(CORBA, __sleep)
{
	CORBA_Environment ev;
	CORBA_Object cobj;
	CORBA_char *ior;
	zval **handle;
	int type;

	CORBA_exception_init(&ev);
	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	cobj = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle), &type);

	phorb_ensure_orb(&ev);

	array_init(return_value);
	ior = CORBA_ORB_object_to_string(OrbSG(orb), cobj, &ev);

	zend_update_property_string(zend_get_class_entry(this_ptr TSRMLS_DC),
	                            this_ptr, "_ior", 4, ior);
	add_next_index_string(return_value, "_ior", 1);

	CORBA_free(ior);
	CORBA_exception_free(&ev);
}

PHP_METHOD(CORBA, __wakeup)
{
	CORBA_Environment ev;
	CORBA_Object cobj;
	zval *handle, **ior;
	int type;

	zend_hash_find(Z_OBJPROP_P(this_ptr), "_ior", 5, (void *)&ior);
	if (!ior) {
		zend_error(E_ERROR, "Bad serialization of CORBA object, "
		                    "no IOR found");
		RETURN_FALSE;
	}
	if (Z_TYPE_PP(ior) != IS_STRING) {
		zend_error(E_ERROR, "Bad serialization of CORBA object, "
		                    "IOR not a string");
		RETURN_FALSE;
	}

	CORBA_exception_init(&ev);
	phorb_ensure_orb(&ev);
	cobj = CORBA_ORB_string_to_object(OrbSG(orb), Z_STRVAL_PP(ior), &ev);
	CORBA_exception_free(&ev);

	if (!cobj) {
		zend_error(E_WARNING, "Could not map IOR to Object while "
		                      "unserializing CORBA Object");
		RETURN_FALSE;
	}

	MAKE_STD_ZVAL(handle);
	ZVAL_RESOURCE(handle, phorb_register_object(cobj));

	zend_hash_index_update(Z_OBJPROP_P(this_ptr), 0, &handle,
	                       sizeof(zval *), NULL);

	RETURN_TRUE;
}

static zend_function_entry class_entries[] = {
	PHP_FE(orbit_load_idl,		arginfo_orbit_load_idl)
	PHP_FE(orbit_exception_handler,	arginfo_orbit_exception_handler)

	PHP_ME(CORBA, __construct,	arginfo_CORBA___construct, 0)
	PHP_ME(CORBA, __call,		arginfo_CORBA___call, 0)
	PHP_ME(CORBA, __get,		arginfo_CORBA___get, 0)
	PHP_ME(CORBA, __set,		arginfo_CORBA___set, 0)
	PHP_ME(CORBA, __sleep,		NULL, 0)
	PHP_ME(CORBA, __wakeup,		NULL, 0)
	PHP_ME(CORBA, _equal,		arginfo_CORBA__equal, 0)

	{ NULL, NULL, NULL }
};

PHP_METHOD(CORBA_TypeCode, __construct)
{
	CORBA_Environment ev;
	CORBA_TypeCode tc;
	char *repoid;
	int repoid_len;
	zval *tmp;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
	                          &repoid, &repoid_len) == FAILURE) {
		zval_dtor(this_ptr);
		ZVAL_NULL(this_ptr);
		RETURN_FALSE;
	}

	tc = phorb_find_typecode(NULL, repoid);
	if (!tc) {
		zend_error(E_WARNING,
		           "Could not find TypeCode for %s in "
		           "CORBA_TypeCode()", repoid);
		zval_dtor(this_ptr);
		ZVAL_NULL(this_ptr);
		RETURN_FALSE;
	}

	MAKE_STD_ZVAL(tmp);
	ZVAL_RESOURCE(tmp, phorb_register_typecode(tc));
	zend_hash_index_update(Z_OBJPROP_P(this_ptr), 0, &tmp,
	                       sizeof(zval *), NULL);
	zend_update_property_string(zend_get_class_entry(this_ptr TSRMLS_DC),
	                            this_ptr, "_repoid", 7, repoid);
}

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_CORBA_TypeCode___construct, 0, 0, 1)
	ZEND_ARG_INFO(0, tcname)
ZEND_END_ARG_INFO()

PHP_METHOD(CORBA_TypeCode, __get)
{
	CORBA_TypeCode tc;
	zval **handle;
	char *name;
	int name_len, type;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
	                          &name, &name_len) == FAILURE) {
		RETURN_FALSE;
	}

	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	tc = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(handle), &type);

	phorb_get_typecode_attribute(tc, name, return_value);
}

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_CORBA_TypeCode___get, 0, 0, 1)
	ZEND_ARG_INFO(0, name)
ZEND_END_ARG_INFO()
											  
PHP_METHOD(CORBA_TypeCode, __set)
{
	CORBA_TypeCode tc;
	zval **handle, *arg;
	char *name;
	int name_len, type;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sz",
	                          &name, &name_len, &arg) == FAILURE) {
		RETURN_FALSE;
	}

	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	tc = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(handle), &type);

	if (phorb_set_typecode_attribute(tc, name, arg) == FAILURE) {
		RETURN_FALSE;
	}
}

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_CORBA_TypeCode___set, 0, 0, 2)
	ZEND_ARG_INFO(0, name)
	ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO()
											  
PHP_METHOD(CORBA_TypeCode, equal)
{
	CORBA_Environment ev;
	CORBA_TypeCode tc, tc2;
	zval **handle, **handle2, *arg;
	int type;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "O",
	                          &arg, *corbatc_class_entry) == FAILURE) {
		RETURN_FALSE;
	}

	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	tc = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(handle), &type);

	zend_hash_index_find(Z_OBJPROP_P(arg), 0, (void *)&handle2);
	tc2 = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(handle2), &type);

	CORBA_exception_init(&ev);
	if (CORBA_TypeCode_equal(tc, tc2, &ev)) {
		RETVAL_TRUE;
	} else {
		RETVAL_FALSE;
	}
	CORBA_exception_free(&ev);
}

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_CORBA_TypeCode_equal, 0, 0, 1)
	ZEND_ARG_INFO(0, typecode2)
ZEND_END_ARG_INFO()

PHP_METHOD(CORBA_TypeCode, __sleep)
{
	CORBA_Environment ev;
	CORBA_TypeCode tc;
	zval **handle;
	char *repoid;
	int type;

	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	tc = (CORBA_TypeCode)zend_list_find(Z_RESVAL_PP(handle), &type);
	array_init(return_value);
	repoid = (char *)tc->repo_id;

	if (!repoid || !repoid[0])
		repoid = "<unknown>";

	zend_update_property_string(zend_get_class_entry(this_ptr TSRMLS_DC),
	                            this_ptr, "_repoid", 7, repoid);
	add_next_index_string(return_value, "_repoid", 1);
}

PHP_METHOD(CORBA_TypeCode, __wakeup)
{
	CORBA_TypeCode tc;
	zval *handle, **repoid;
	int type;

	zend_hash_find(Z_OBJPROP_P(this_ptr), "_repoid", 8, (void *)&repoid);
	if (!repoid) {
		zend_error(E_ERROR, "Bad serialization of CORBA_TypeCode "
		                    "object, no repoid found");
		RETURN_FALSE;
	}
	if (Z_TYPE_PP(repoid) != IS_STRING) {
		zend_error(E_ERROR, "Bad serialization of CORBA_TypeCode "
		                    "object, repoid not a string");
		RETURN_FALSE;
	}

	tc = phorb_find_typecode(NULL, Z_STRVAL_PP(repoid));
        if (!tc) {
		zend_error(E_WARNING, "Could not find repoid `%s' in "
		                      "typecodes", Z_STRVAL_PP(repoid));
		RETURN_FALSE;
	}

	MAKE_STD_ZVAL(handle);
	ZVAL_RESOURCE(handle, phorb_register_typecode(tc));

	zend_hash_index_update(Z_OBJPROP_P(this_ptr), 0, &handle,
	                       sizeof(zval *), NULL);

	RETURN_TRUE;
}

static zend_function_entry tcclass_entries[] = {
	PHP_ME(CORBA_TypeCode, __construct,	arginfo_CORBA_TypeCode___construct, 0)
	PHP_ME(CORBA_TypeCode, __get,		arginfo_CORBA_TypeCode___get, 0)
	PHP_ME(CORBA_TypeCode, __set,		arginfo_CORBA_TypeCode___set, 0)
	PHP_ME(CORBA_TypeCode, __sleep,		NULL, 0)
	PHP_ME(CORBA_TypeCode, __wakeup,	NULL, 0)
	PHP_ME(CORBA_TypeCode, equal,		arginfo_CORBA_TypeCode_equal, 0)

	{ NULL, NULL, NULL }
};

int phorb_register_destructor(void (*simple)(void *),
                              void (*persistent)(void *))
{
	return zend_register_list_destructors(simple, persistent, mod_number);
}

static void php_orbit_init_globals(zend_orbit_globals *orbit_globals)
{
	memset(orbit_globals, 0, sizeof orbit_globals);
}

PHP_INI_BEGIN()
	STD_PHP_INI_ENTRY("orbit.idl_cache_timeout", "300", PHP_INI_SYSTEM, OnUpdateLong, idl_cache_timeout, zend_orbit_globals, orbit_globals)
	STD_PHP_INI_ENTRY("orbit.idl_cache_check", "30", PHP_INI_SYSTEM, OnUpdateLong, idl_cache_check, zend_orbit_globals, orbit_globals)
	STD_PHP_INI_ENTRY("orbit.idl_cpp_arguments", "-D__ORBIT_IDL__ -iquote.", PHP_INI_ALL, OnUpdateString, idl_cpp_arguments, zend_orbit_globals, orbit_globals)
PHP_INI_END()

PHP_MINIT_FUNCTION(orbit)
{
	zend_class_entry corba_class;
	zend_class_entry corbatc_class;

	ZEND_INIT_MODULE_GLOBALS(orbit, php_orbit_init_globals, NULL);
	REGISTER_INI_ENTRIES();
	mod_number = module_number;

	INIT_CLASS_ENTRY(corba_class, "CORBA", class_entries);
	corba_class_entry =
		zend_register_internal_class(&corba_class TSRMLS_CC);
	zend_declare_property_string(corba_class_entry, "_ior",
	                             4, "", ZEND_ACC_PUBLIC TSRMLS_CC);

	INIT_CLASS_ENTRY(corbatc_class, "CORBA_TypeCode", tcclass_entries);
	corbatc_class_entry =
		zend_register_internal_class(&corbatc_class TSRMLS_CC);
	zend_declare_property_string(corbatc_class_entry, "_repoid",
	                             7, "", ZEND_ACC_PUBLIC TSRMLS_CC);

	phorb_init_corba_class(corba_class_entry);
	phorb_init_typecode_class(corbatc_class_entry);

	OrbSG(idl_cache) =
		g_hash_table_new_full(g_str_hash, g_str_equal,
		                      NULL, (GDestroyNotify)phorb_free_idl);

	OrbSG(idls) = NULL;
	OrbSG(orb) = NULL;

	return SUCCESS;
}

PHP_MSHUTDOWN_FUNCTION(orbit)
{
	phorb_shutdown_orb();

	g_hash_table_destroy(OrbSG(idl_cache));

	UNREGISTER_INI_ENTRIES();

	phorb_fini_corba_class();
	phorb_fini_typecode_class();

	return SUCCESS;	
}

PHP_RINIT_FUNCTION(orbit)
{
	phorb_check_idl_cache();

	phorb_init_exceptions();

	return SUCCESS;
}

PHP_RSHUTDOWN_FUNCTION(orbit)
{
	GSList *list;

	phorb_shutdown_exceptions();

	for(list = OrbSG(idls); list; list = list->next) {
		phorbIDL *idl = (phorbIDL *)list->data;

		idl->lastaccessed = time(NULL);
	}
	g_slist_free(OrbSG(idls));
	OrbSG(idls) = NULL;

	OrbSG(active) = 0;
	if (!OrbSG(counter))
		phorb_shutdown_orb();

	return SUCCESS;
}

PHP_MINFO_FUNCTION(orbit)
{
	char buf[32];

	php_info_print_table_start();
	php_info_print_table_header(2, "ORBit Support", "enabled");
	php_info_print_table_end();
	DISPLAY_INI_ENTRIES();
}

zend_module_entry orbit_module_entry = {
	STANDARD_MODULE_HEADER,
	PHP_ORBIT_EXTNAME,
	class_entries,
	PHP_MINIT(orbit),
	PHP_MSHUTDOWN(orbit),
	PHP_RINIT(orbit),
	PHP_RSHUTDOWN(orbit),
	PHP_MINFO(orbit),
	PHP_ORBIT_VERSION,
	STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_ORBIT
ZEND_GET_MODULE(orbit)
#endif
