#include "php_orbit.h"
#include "phorb.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CORBA_dup(t, v)	(t)CORBA_Object_duplicate((CORBA_Object)(v), NULL)

static zend_class_entry *corbatc_class;

void phorb_init_typecode_class(zend_class_entry *tcclass)
{
	corbatc_class = tcclass;
}

void phorb_fini_typecode_class(void)
{
}

gboolean phorb_is_typecode(zval *val)
{
	return Z_TYPE_P(val) == IS_OBJECT && Z_OBJCE_P(val) == corbatc_class;
}

zval *phorb_init_typecode(zval *res)
{
	object_init_ex(res, corbatc_class);

	return res;
}

ulong phorb_register_typecode(CORBA_TypeCode tc)
{
	return phorb_register_object((CORBA_Object)tc);
}

CORBA_TypeCode phorb_find_typecode(phorbIDL *idl, const char *repoid)
{
	GSList *list;
	CORBA_TypeCode ret = CORBA_OBJECT_NIL;

	if (idl)
		ret = g_hash_table_lookup(idl->typecodes, repoid);
	else {
		for(list = OrbSG(idls); list; list = list->next) {
			idl = (phorbIDL *)list->data;
			ret = g_hash_table_lookup(idl->typecodes, repoid);
			if (ret)
				break;
		}
	}

	return ret ? CORBA_dup(CORBA_TypeCode, ret) : CORBA_OBJECT_NIL;
}

void phorb_store_typecode(phorbIDL *idl, CORBA_TypeCode tc)
{
	tc = CORBA_dup(CORBA_TypeCode, tc);
	g_hash_table_insert(idl->typecodes, tc->repo_id, tc);
}

#define ADD_BASIC(type)	phorb_store_typecode(idl, TC_CORBA_##type)

void phorb_setup_typecodes(phorbIDL *idl)
{
	ADD_BASIC(short);
	ADD_BASIC(long);
	ADD_BASIC(long_long);
	ADD_BASIC(unsigned_short);
	ADD_BASIC(unsigned_long);
	ADD_BASIC(unsigned_long_long);
	ADD_BASIC(float);
	ADD_BASIC(double);
	ADD_BASIC(long_double);
	ADD_BASIC(boolean);
	ADD_BASIC(char);
	ADD_BASIC(wchar);
	ADD_BASIC(octet);
	ADD_BASIC(any);
	ADD_BASIC(TypeCode);
	ADD_BASIC(Object);
	ADD_BASIC(string);
	ADD_BASIC(wstring);
}

#undef ADD_BASIC

void phorb_get_typecode_attribute(CORBA_TypeCode tc, const char *name, zval *return_value)
{
	if (strcmp(name, "kind") == 0) {
		ZVAL_STRING(return_value,
		            (char *)ORBit_tk_to_name(tc->kind), 1);
	} else if (strcmp(name, "flags") == 0) {
		ZVAL_LONG(return_value, tc->flags);
	} else if (strcmp(name, "length") == 0) {
		ZVAL_LONG(return_value, tc->length);
	} else if (strcmp(name, "sub_parts") == 0) {
		ZVAL_LONG(return_value, tc->sub_parts);
	} else if (strcmp(name, "subtypes") == 0) {
		int i;

		if (tc->kind != CORBA_tk_struct
		    && tc->kind != CORBA_tk_except
		    && tc->kind != CORBA_tk_union
		    && tc->kind != CORBA_tk_alias
		    && tc->kind != CORBA_tk_array
		    && tc->kind != CORBA_tk_sequence) {
			zend_error(E_WARNING, "subtypes not available for "
			                      "this type");
			RETURN_FALSE;
		}

		array_init(return_value);
		for(i = 0; i < tc->sub_parts; i++) {
			CORBA_TypeCode subtc = tc->subtypes[i];
			zval *obj, *val;

			MAKE_STD_ZVAL(obj);
			obj = phorb_init_typecode(obj);
			if (!obj) {
				zval_dtor(obj);
				FREE_ZVAL(obj);
				RETURN_FALSE;
			}

			MAKE_STD_ZVAL(val);
			ZVAL_RESOURCE(val, phorb_register_typecode(
				(CORBA_TypeCode)CORBA_Object_duplicate(
					(CORBA_Object)subtc, NULL)));
			zend_hash_index_update(Z_OBJPROP_P(obj), 0, &val,
			                       sizeof(zval *), NULL);

			add_next_index_zval(return_value, obj);
		}
	} else if (strcmp(name, "discriminator") == 0) {
		zval *val;

		if (tc->kind != CORBA_tk_union) {
			zend_error(E_WARNING, "discriminator not available "
			                      "for this type");
			RETURN_FALSE;
		}

		return_value = phorb_init_typecode(return_value);
		if (!return_value)
			RETURN_FALSE;

		MAKE_STD_ZVAL(val);
		ZVAL_RESOURCE(val, phorb_register_typecode(
			(CORBA_TypeCode)CORBA_Object_duplicate(
		        	(CORBA_Object)tc->discriminator, NULL)));
		zend_hash_index_update(Z_OBJPROP_P(return_value), 0, &val,
		                       sizeof(zval *), NULL);
	} else if (strcmp(name, "name") == 0) {
		ZVAL_STRING(return_value, tc->name, 1);
	} else if (strcmp(name, "repo_id") == 0) {
		ZVAL_STRING(return_value, tc->repo_id, 1);
	} else if (strcmp(name, "subnames") == 0) {
		int i;

		if (tc->kind != CORBA_tk_struct
		    && tc->kind != CORBA_tk_except
		    && tc->kind != CORBA_tk_union
		    && tc->kind != CORBA_tk_enum) {
			zend_error(E_WARNING, "subnames not available for "
			                      "this type");
			RETURN_FALSE;
		}

		array_init(return_value);
		for(i = 0; i < tc->sub_parts; i++)
			add_next_index_string(return_value, tc->subnames[i], 1);
	} else if (strcmp(name, "sublabels") == 0) {
		int i;

		if (tc->kind != CORBA_tk_union) {
			zend_error(E_WARNING, "sublabels not available for "
			                      "this type");
			RETURN_FALSE;
		}

		array_init(return_value);
		for(i = 0; i < tc->sub_parts; i++)
			add_next_index_long(return_value, tc->sublabels[i]);
	} else if (strcmp(name, "default_index") == 0) {
		if (tc->kind != CORBA_tk_union) {
			zend_error(E_WARNING, "default_index not available "
			                      "for this type");
			RETURN_FALSE;
		}

		ZVAL_LONG(return_value, tc->default_index);
	} else if (strcmp(name, "recurse_depth") == 0) {
		if (tc->kind != CORBA_tk_sequence) {
			zend_error(E_WARNING, "recurse_depth not available "
			                      "for this type");
			RETURN_FALSE;
		}

		ZVAL_LONG(return_value, tc->recurse_depth);
	} else if (strcmp(name, "digits") == 0) {
		if (tc->kind != CORBA_tk_fixed) {
			zend_error(E_WARNING, "digits not available for"
			                      "this type");
			RETURN_FALSE;
		}

		ZVAL_LONG(return_value, tc->digits);
	} else if (strcmp(name, "scale") == 0) {
		if (tc->kind != CORBA_tk_fixed) {
			zend_error(E_WARNING, "scale not available for "
			                      "this type");
			RETURN_FALSE;
		}

		ZVAL_LONG(return_value, tc->scale);
	} else if (strcmp(name, "_hash") == 0) {
		CORBA_Environment ev;

		CORBA_exception_init(&ev);
		ZVAL_LONG(return_value, CORBA_Object_hash((CORBA_Object)tc,
		                                          LONG_MAX, &ev));
		CORBA_exception_free(&ev);
	} else {
		zend_error(E_WARNING, "Non-existing attribute `%s'",
		           name);
	}
}

int phorb_set_typecode_attribute(CORBA_TypeCode tc, const char *name, zval *value)
{
	zend_error(E_WARNING, "Typecode attributes are readonly");
	return FAILURE;
}

CORBA_long phorb_enum_find_member(CORBA_TypeCode tc, zval *val)
{
	char *str;
	CORBA_long i;

	convert_to_string(val);
	str = Z_STRVAL_P(val);
	if (!str)
		return -1;

	for(i = 0; i < tc->sub_parts; i++)
		if (strcmp(tc->subnames[i], str) == 0)
			return i;

	return -1;
}

#define ZVAL_RES_FIND_MEMBER(kind, type) \
	case CORBA_tk_##kind: do { \
		CORBA_##type val = Z_RESVAL_P(discriminator); \
		\
		for(i = 0; i < tc->sub_parts; i++) { \
			CORBA_##type label = \
				*(CORBA_##type *)tc->sublabels[i]; \
			\
			if (label == val) \
				return i; \
		} \
		break; \
	} while(0)

CORBA_long phorb_union_find_arm(CORBA_TypeCode tc, zval *discriminator)
{
	CORBA_unsigned_long i;

	/* assuming discriminator IS_RESOURCE */

	switch(tc->discriminator->kind) {
	ZVAL_RES_FIND_MEMBER(char, char);
	ZVAL_RES_FIND_MEMBER(octet, octet);
	ZVAL_RES_FIND_MEMBER(short, short);
	ZVAL_RES_FIND_MEMBER(ushort, unsigned_short);
	ZVAL_RES_FIND_MEMBER(long, long);
	ZVAL_RES_FIND_MEMBER(ulong, unsigned_long);
	ZVAL_RES_FIND_MEMBER(longlong, long_long);
	ZVAL_RES_FIND_MEMBER(ulonglong, unsigned_long_long);
	case CORBA_tk_enum: {
		CORBA_unsigned_long val =
			phorb_enum_find_member(
				tc->discriminator, discriminator);

		for(i = 0; i < tc->sub_parts; i++)
			if (tc->sublabels[i] == val)
				return i;
		break;
	}
	case CORBA_tk_boolean: {
		CORBA_boolean val = Z_RESVAL_P(discriminator);

		for(i = 0; i < tc->sub_parts; i++)
			if (!tc->sublabels[i] == !val)
				return i;
		break;
	}
	default:
		zend_error(E_WARNING,
		           "Unsupported discriminator type %d",
		           tc->discriminator->kind);
	}

	return (tc->default_index >= 0) ? tc->default_index : -1;
}

#undef ZVAL_RES_FIND_MEMBER
